import numpy as np
import qiskit.quantum_info as qi

class RBS:
    """
    Implementation of a RBS gate as an operator.
    """
    def __init__(self, theta: float) -> None:
        """Create new RBS operation."""
        a = np.cos(theta)
        b = np.sin(theta)

        # Matrix definition of the gate
        self.rbs = qi.Operator([ [1, 0,  0, 0],
                                 [0, a,  b, 0],
                                 [0, -b, a, 0],
                                 [0, 0,  0, 1]])
        
    @property
    def rbs(self):
        return self._rbs
    
    @rbs.setter
    def rbs(self, value):
        self._rbs = value


"""
=================================================================
Quantum image classifier module (:mod:`quantum_image_classifier`)
=================================================================
.. currentmodule:: quantum_image_classifier

This is the quantum image classifier implemented by Jorge Vázquez Pérez. The set of algorithms here
implemented can be used with :mod:`~quantum_image_classifier.NAME_OF_THE_ALGORITHM`.
"""

from .classifier_algorithms import NearestCentroid
from .data_treatment import data_generator, data_loader, QuantumAE, QuantumAE2
from .encoding import UnaryLoader, PhaseEncoding
from .gates import RBS, iRBS
from .visuals import error_graph, variance_error_graph


from optparse import OptionError
import pandas as pd
import numpy as np
import os
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from keras.layers import Dense,Conv2D,MaxPooling2D,UpSampling2D,Flatten, Reshape
from keras import Sequential, Model, Input
import tensorflow.keras as tk
import numpy as np
import matplotlib.pyplot as plt

import os
dirname = os.path.dirname(__file__)
train = os.path.join(dirname, '../../data/mnist_train.csv')
test = os.path.join(dirname, '../../data/mnist_test.csv')

def get_MNIST(n_components, reduction: str = "PCA", labels: np.ndarray = [0,1,2,3,4,5,6,7,8,9]) -> tuple:
    """
    Function to get the MNIST dataset and perform a dimension reduction to it.
    To perform the dimension reduction we can use:
        - PCA
        - Simple autoencoder
        - Autoencoder based on CNN.

    Args:
        n_components: number of components to which we reduce the data
        reduction: type of reduction we will use
    """
    mnist = tk.datasets.mnist
    (train_X, train_y), (test_X, test_y) = mnist.load_data()

    if not labels == [0,1,2,3,4,5,6,7,8,9]:
        train_mask = np.isin(train_y, labels)
        test_mask = np.isin(test_y, labels)
        train_X, train_y = train_X[train_mask], train_y[train_mask]
        test_X, test_y = test_X[test_mask], test_y[test_mask]

    if reduction == "PCA":
        # We apply PCA to each the training and the test dataset
        train_X, test_X = _do_PCA(n_components, train_X, test_X)
    elif reduction == "AE":
        train_X, test_X = _do_AE(n_components, train_X, test_X)
    elif reduction == "AE_CNN":
        train_X, test_X = _do_AE_CNN(n_components, train_X, test_X)
    elif reduction == "None":
        train_X = train_X.reshape((len(train_X), 784))
        test_X = test_X.reshape((len(test_X), 784))
        return train_X, train_y, test_X, test_y
    else:
        raise OptionError()

    return train_X, train_y, test_X, test_y


def _do_PCA(n_components: int, train_X: np.ndarray, test_X: np.ndarray) -> tuple:
    """
    Function to perform PCA to a given data.

    Args:
        n_components: number of components to which we reduce the data
        data: the info we want to apply PCA to
    Returns:
        tuple: with the train and test arrays preprocessed
    """

    # We reshape and standarize the data
    train_X = train_X.reshape((len(train_X), 784))
    test_X = test_X.reshape((len(test_X), 784))
    train_X = StandardScaler().fit_transform(train_X)
    test_X = StandardScaler().fit_transform(test_X)
    pca = PCA(n_components)

    # We perform the PCA
    train_X = pca.fit_transform(train_X)
    test_X = pca.fit_transform(test_X)

    return train_X, test_X

def _do_AE(encoding_dim: int, train_X: np.ndarray, test_X: np.ndarray) -> tuple:
    """
    Function to apply a simple autoencoder to a given data.

    Args:
        n_components: number of components to which we reduce the data
        data: the info we want to apply the simple autoencoder to
    Returns:
        tuple: with the train and test arrays preprocessed
    """
    train_X = train_X.reshape((len(train_X), 784))
    test_X = test_X.reshape((len(test_X), 784))

    # Autoencoder model creation
    input_img = Input(shape=(784,))
    encoded = Dense(encoding_dim, activation='relu')(input_img)
    decoded = Dense(784, activation='sigmoid')(encoded)
    autoencoder = Model(input_img, decoded)

    # Encoder part constructed
    encoder = Model(input_img, encoded)

    # Compile of the model, scale the data and fit of the model
    autoencoder.compile(optimizer='adam', loss='binary_crossentropy')
    train_X = train_X.astype('float32') / 255.
    test_X = test_X.astype('float32') / 255.
    autoencoder.fit(train_X, train_X,
                epochs=20,
                batch_size=256,
                validation_data=(test_X, test_X))

    
    return encoder.predict(train_X), encoder.predict(test_X)

def _do_AE_CNN(n_components: int, train_X: np.ndarray, test_X: np.ndarray) -> tuple:
    """
    Function to apply a autoencoder based on a CNN to a given data.

    Args:
        n_components: number of components to which we reduce the data
        data: the info we want to apply the autoencoder based on a CNN to
    Returns:
        tuple: with the train and test arrays preprocessed
    """

    model = Sequential()

    # Encoder network
    model.add(Conv2D(30, 3, activation= 'relu', padding='same', input_shape = (28,28,1)))
    model.add(MaxPooling2D(2, padding= 'same'))
    model.add(Conv2D(15, 3, activation= 'relu', padding='same'))
    model.add(MaxPooling2D(2, padding= 'same'))
    model.add(Flatten())
    model.add(Dense(units=n_components, activation="relu"))

    # Decoder network
    model.add(Dense(units=735, activation="relu"))
    model.add(Reshape((7, 7, 15)))
    model.add(Conv2D(15, 3, activation= 'relu', padding='same'))
    model.add(UpSampling2D(2))
    model.add(Conv2D(30, 3, activation= 'relu', padding='same'))
    model.add(UpSampling2D(2))
    model.add(Conv2D(1,3,activation='sigmoid', padding= 'same')) # output layer
    model.compile(optimizer= 'adam', loss = 'binary_crossentropy')

    # Scalate the data and fit the model
    train_X = train_X.astype('float32') / 255.
    test_X = test_X.astype('float32') / 255.
    model.fit(train_X, train_X, epochs=8, batch_size=128, validation_data=(test_X, test_X))
    
    # Create new model to get the data encoded
    new_model= Sequential()
    new_model.add(Conv2D(30, 3, activation= 'relu', padding='same', input_shape = (28,28,1)))
    new_model.add(MaxPooling2D(2, padding= 'same'))
    new_model.add(Conv2D(15, 3, activation= 'relu', padding='same'))
    new_model.add(MaxPooling2D(2, padding= 'same'))
    new_model.add(Flatten())
    new_model.add(Dense(units=n_components, activation="relu"))

    # Set weights of the trained autoencoder to the encoder model
    enc_len = len(model.layers) // 2
    for i, layer in enumerate(model.layers[:enc_len]):
        new_model.layers[i].set_weights(layer.get_weights())

    # Compile it after setting the weights
    new_model.compile(optimizer='adam', loss='categorical_crossentropy')
    
    return new_model.predict(train_X), new_model.predict(test_X)







import cmath, math
import numpy as np

from qiskit import QuantumCircuit, QuantumRegister

from .encoding import Encoding


class PhaseEncoding(Encoding):

    def __init__(self, input_vector: np.ndarray):
        input_vector = (input_vector.astype(float) / 255)* math.pi
        self._generate_circuit(input_vector)
        

        super().__init__("Phase encoding")
        

    def _generate_circuit(self, input_vector: np.ndarray):
        self.num_qubits = math.ceil(math.log2(len(input_vector)))

        self.quantum_data = QuantumRegister(self.num_qubits)
        self.circuit = QuantumCircuit(self.quantum_data)

        initial_state = np.array([], dtype=np.complex_)
        for k in range(len(input_vector)):
            x_k = complex(0, input_vector[k])
            initial_state = np.append(initial_state, cmath.exp(x_k))

        norm = 0
        for elem in initial_state:
            norm += elem * np.conj(elem)
        initial_state = initial_state / math.sqrt(norm)

        for _ in range(2**self.num_qubits - len(input_vector)):
            initial_state = np.append(initial_state, 0)
        
        self.circuit.initialize(initial_state, self.quantum_data)


import setuptools
import inspect
import sys
import os
import re

REQUIREMENTS_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), "env",  "requirements.txt")
with open(REQUIREMENTS_PATH) as f:
    REQUIREMENTS = f.read().splitlines()

if not hasattr(setuptools, 'find_namespace_packages') or not inspect.ismethod(setuptools.find_namespace_packages):
    print("Your setuptools version:'{}' does not support PEP 420 (find_namespace_packages). "
          "Upgrade it to version >='40.1.0' and repeat install.".format(setuptools.__version__))
    sys.exit(1)

VERSION_PATH = os.path.join(os.path.dirname(__file__), "VERSION.txt")
with open(VERSION_PATH, "r") as version_file:
    VERSION = version_file.read().strip()

# Read long description from README.
README_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")
with open(README_PATH) as readme_file:
    README = re.sub(
        "<!--- long-description-skip-begin -->.*<!--- long-description-skip-end -->",
        "",
        readme_file.read(),
        flags=re.S | re.M,
    )


setuptools.setup(
    name='quantum-image-classifier',
    version=VERSION,
    description='Quatum image classifier: A library of different quantum algorithms used to classify images',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/jorgevazquezperez/Quantum-image-classifier',
    author='Jorge Vázquez Pérez',
    author_email='jorge.vazper@gmail.com',
    #license='Apache-2.0',
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering"
    ],
    keywords='qiskit quantum machine learning ml centroids',
    packages=[
        "quantum_image_classifier", 
        "quantum_image_classifier.classifier_algorithms", 
        "quantum_image_classifier.encoding", 
        "quantum_image_classifier.gates",
        "quantum_image_classifier.data_treatment" ],
    install_requires=REQUIREMENTS,
    include_package_data=True,
    python_requires=">=3.7",
    zip_safe=False
)


import numpy as np

from qiskit.circuit import ParameterVector
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, Aer, assemble, transpile
from qiskit.algorithms.optimizers import ADAM

from scipy.optimize import basinhopping

from ..encoding import PhaseEncoding


class QuantumAE():
    def __init__(self, num_qubits: int, n_components: int):
        self.num_qubits = num_qubits
        self.n_trash = num_qubits - n_components
        self._build_autoencoder()

    def _build_autoencoder(self):

        self._build_circuit()
        self._build_SWAPTest(self.n_trash)

        self.qregisters = QuantumRegister(
            self.num_qubits + self.n_trash + 1, "q")
        self.autoencoder = QuantumCircuit(self.qregisters, name="Autoencoder")
        self.autoencoder.compose(
            self.circuit, qubits=self.qregisters[self.n_trash + 1:], inplace=True)
        self.autoencoder.compose(
            self.swap_test, qubits=self.qregisters[:2*self.n_trash + 1], inplace=True)

    def _build_circuit(self):
        N_PARAMETERS = self.num_qubits*4 + \
            self.num_qubits*(self.num_qubits - 1)
        self.params = ParameterVector('θ', N_PARAMETERS)

        self.circuit = QuantumCircuit(self.num_qubits, name="Main circuit")
        self.circuit.barrier()

        weight_counter = 0

        for i in range(self.num_qubits):
            self.circuit.rx(self.params[weight_counter], i)
            self.circuit.rz(self.params[weight_counter+1], i)
            weight_counter += 2

        iterator = list(range(self.num_qubits-1, -1, -1))
        for i in iterator:
            iterator_j = iterator.copy()
            iterator_j.remove(i)
            for j in iterator_j:
                self.circuit.crx(self.params[weight_counter], i, j)
                weight_counter += 1

        for i in range(self.num_qubits):
            self.circuit.rx(self.params[weight_counter], i)
            self.circuit.rz(self.params[weight_counter+1], i)
            weight_counter += 2

    def _build_SWAPTest(self, n_trash):
        qregisters = QuantumRegister(2*n_trash + 1, "q")
        self.swap_test = QuantumCircuit(qregisters, name="SWAP test")
        self.swap_test.barrier()

        self.swap_test.h(qregisters[0])
        for i in range(n_trash):
            self.swap_test.cswap(
                qregisters[0], qregisters[i+1], qregisters[n_trash+i+1])
        self.swap_test.h(qregisters[0])

    def fit(self, dataset: np.ndarray = None) -> None:

        opt = ADAM(lr=0.1, maxiter=200)

        pt = np.random.uniform(low=0,
                               high=4*np.pi,
                               size=(len(self.params),))

        self.states = []
        for point in dataset:
            self.states.append(PhaseEncoding(point))

        print(self.states)
        def loss(params):
            expectation = 0
            dict = {}
            for param, value in zip(self.params, params):
                dict[param] = value
            for state in self.states:
                cregister = ClassicalRegister(1, "c")
                circ_aux = QuantumCircuit(self.qregisters, cregister)
                circ_aux.compose(
                    state.circuit, qubits=self.qregisters[self.n_trash + 1:], inplace=True)
                circ_aux.compose(self.autoencoder,
                                 qubits=self.qregisters, inplace=True)
                circ_aux.measure(self.qregisters[0], cregister[0])

                simulator = Aer.get_backend('aer_simulator')
                t_qc = transpile(circ_aux,
                                 simulator)

                qobj = assemble(t_qc,
                                shots=1024,
                                parameter_binds=[dict])
                job = simulator.run(qobj)
                result = job.result().get_counts()

                counts = np.array(list(result.values()))
                states = np.array(list(result.keys())).astype(float)
                probabilities = counts / 1024
                expectation += np.sum(states * probabilities)

            return expectation / len(dataset)

        """# perform the basin hopping search
        result = basinhopping(loss, pt, stepsize=0.06, niter=20, disp=True)
        # summarize the result
        print('Status : %s' % result['message'])
        print('Total Evaluations: %d' % result['nfev'])
        # evaluate solution
        solution = result['x']
        evaluation = loss(solution)
        print('Solution: f(%s) = %.5f' % (solution, evaluation))"""

        result = opt.minimize(loss, pt)
        # summarize the result
        print('Total Evaluations: %d' % result.nfev)
        # evaluate solution
        solution = result.x
        evaluation = result.fun
        print('Solution: f(%s) = %.5f' % (solution, evaluation))
